/*  Copyright (C) 2005-2007 Tresys Technology, LLC
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * Author: jmowery@tresys.com
 *
 */

#ifndef SECHK_PARSE_H
#define SECHK_PARSE_H

#ifdef	__cplusplus
extern "C"
{
#endif

#include "sechecker.h"

#include <libxml/xmlreader.h>
#include <assert.h>

#define sechk_lib_parse_profile(path, lib)     sechk_lib_parse_xml_file(path, lib)

	int sechk_lib_parse_xml_file(const char *filename, sechk_lib_t * lib);
	int sechk_lib_process_xml_node(xmlTextReaderPtr reader, sechk_lib_t * lib);

#ifdef	__cplusplus
}
#endif

#endif
