# seaudit-report.css
#
# This is a default stylesheet template used to format an HTML report
# generated by the seaudit-report tool. 
#
# There are two ways of configuring this style sheet:
#
# 1. External Style Sheet:
# 	Specify an external stylesheet to link to the HTML report using a 
# 	<link> tag, as follows:
# 		<link rel="stylesheet" type="text/css" href="mystyle.css">
# 	The seaudit-report tool will insert the <link> tag you define 
#	within the <head> section of the generated HTML report.
#
# 2. Internal Style Sheet:
#	Specify internal style tags for the HTML report using the <style> 
# 	tag. An example can be:
# 		<style type="text/css">
#		body {background-color: red}
#		p {margin-left: 20px}
#		</style>
# 	The seaudit-report tool will insert the <style> tag you define 
#	within the head section of the generated HTML report.
#
# The following is a list of defined HTML tags and their class selectors
# that can be used:
# 	
# HTML Tag		Class Selector(s)
# -------- 		-----------------
# <h1>			report_title
#
# <h2>			standard_section_title, custom_section_title
#
# <font>		message_count_label, stats_label, message_date,
#			host_name, syscall_timestamp, avc_type,
#			src_context, tgt_context, obj_class
#
# <b>			report_date, message_count, stats_count
#
# See the default (internal) style sheet format below for an example.s

body
{
	font-size: 82%;
	color:#000000;
	background-color:white;
	margin:0px;
}

h1.report_title
{
	margin-top:0px;
	margin-bottom:5px;
	font-size:110%;
	padding-top:0px;
	padding-bottom:1px;
	padding-left:4px;
	color:#ffffff;
	text-align:left;
	background-color:#808080;
}

h2.standard_section_title
{
	margin-top:0px;
	margin-bottom:5px;
	font-size:100%;
	padding-top:0px;
	padding-bottom:1px;
	padding-left:0px;
	color:black;
	background-color:transparent;
}

h2.custom_section_title
{
	margin-top:0px;
	margin-bottom:5px;
	font-size:100%;
	padding-top:0px;
	padding-bottom:1px;
	padding-left:0px;
	color:black;
	background-color:transparent;
}

font.message_count_label
{
color: #000000;
background-color:transparent;
}

b.message_count
{
color: #000000;
background-color:transparent;
}

font.stats_label
{
color: #000000;
background-color:transparent;
}

b.stats_count
{
color: #000000;
background-color:transparent;
}

b.report_date
{
color: #000000;
background-color:transparent;
}

font.message_date
{
color: black;
background-color:transparent;
position:relative
}

font.host_name
{
color: #000000;
background-color:transparent;
}

font.syscall_timestamp
{
color: #000000;
background-color:transparent;
}

font.avc_deny
{
color:red;
background-color:transparent;
}

font.avc_grant
{
color:green;
background-color:transparent;
}

font.exe
{
color: #000000;
background-color:transparent;
}

font.path
{
color: blue;
background-color:transparent;
}

font.src_context
{
color:black;
background-color:transparent;
font-weight: bold
}

font.tgt_context
{
color:black;
background-color:transparent;
font-weight: bold
}

font.obj_class
{
color: #000000;
background-color:transparent;
}